#include "common.h"
#include <atomic>
#include <cds/gc/dhp.h>
#include <cds/gc/hp.h>
#include <cds/sync/barrier.h>
#include <cds_test/stress_test.h>
#include <iostream>
#include <memory>
#include <thread>

using namespace std;

namespace {

typedef cds_others::SpinBarrier Barrier;

static size_t s_nBarrierThreadCount = 6;
static size_t s_nBarrierPassCount = 100000000;

class BarrierTest : public cds_test::stress_fixture {
protected:
  static Barrier *barrier;
  static size_t count;

  static void SetUpTestCase() {
    cds_test::config const &cfg = get_config("Misc");
    GetConfig(BarrierPassCount);
    GetConfig(BarrierThreadCount);
  }

  static void TearDownTestCase() {
    if (count != s_nBarrierPassCount) {
      cout << "Incorrect" << endl;
    }
  }

  static void Thread() {
    for (ullong i = 0; i < s_nBarrierPassCount; i++) {
      if (barrier->wait()) {
        count++;
      }
    }
  }
};

Barrier *BarrierTest::barrier;
size_t BarrierTest::count;

TEST_F(BarrierTest, Wait) {
  barrier = new Barrier(s_nBarrierThreadCount);
  std::unique_ptr<std::thread[]> threads(new std::thread[s_nBarrierThreadCount]);
  for (size_t i = 0; i < s_nBarrierThreadCount; i++) {
    threads[i] = std::thread(Thread);
  }

  for (size_t i = 0; i < s_nBarrierThreadCount; i++) {
    threads[i].join();
  }
}

} // namespace
