#include "common.h"
#include <atomic>
#include <cds/gc/dhp.h>
#include <cds/gc/hp.h>
#include <cds/sync/seqlock.h>
#include <cds_test/stress_test.h>
#include <iostream>
#include <memory>
#include <thread>

using namespace std;

namespace {

typedef cds_others::SeqLock SeqLock;

static size_t s_nSeqLockReaderWriterThreadCount = 6;
static size_t s_nSeqLockPassCount = 2000000;

class SeqLockTest : public cds_test::stress_fixture {
protected:
  static size_t sum;
  static SeqLock *seqlock;

  static void SetUpTestCase() {
    cds_test::config const &cfg = get_config("Misc");
    GetConfig(SeqLockReaderWriterThreadCount);
    GetConfig(SeqLockPassCount);
  }

  static void ReaderWriterThread(int write_percentage) {
    for (size_t i = 0; i < s_nSeqLockPassCount; i++) {
      if (rand(100) < write_percentage) {
        sum += seqlock->read();
      } else {
        seqlock->write(rand(10));
      }
    }
  }
};

size_t SeqLockTest::sum;
SeqLock *SeqLockTest::seqlock;

TEST_F(SeqLockTest, BasicReadWriter) {
  seqlock = new SeqLock();
  for (int write_percentage = 5; write_percentage < 50; write_percentage += 5) {
    std::unique_ptr<std::thread[]> threads(
        new std::thread[s_nSeqLockReaderWriterThreadCount]);
    for (size_t i = 0; i < s_nSeqLockReaderWriterThreadCount; i++) {
      threads[i] = std::thread(ReaderWriterThread, write_percentage);
    }
    for (size_t i = 0; i < s_nSeqLockReaderWriterThreadCount; i++) {
      threads[i].join();
    }
  }
}

} // namespace
