/*
    This file is a part of libcds - Concurrent Data Structures library

    (C) Copyright Maxim Khizhinsky (libcds.dev@gmail.com) 2006-2017

    Source code repo: http://github.com/khizmax/libcds/
    Download: http://sourceforge.net/projects/libcds/files/

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
   this
    list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
   ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
   LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
   USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "../misc/common.h"
#include "../stack/stack_type.h"

namespace {

#define CDSSTRESS_SequentialTreiberStack_F(test_fixture, type_name)            \
  TEST_F(test_fixture, type_name) {                                            \
    typedef stack::Types<value_type>::type_name stack_type;                    \
    stack_type stack;                                                          \
    test(stack);                                                               \
  }

#define CDSSTRESS_SequentialEliminationStack_F(test_fixture, type_name)        \
  TEST_F(test_fixture, type_name) {                                            \
    typedef stack::Types<value_type>::type_name stack_type;                    \
    stack_type stack(s_nSequentialEliminationSize);                            \
    test(stack);                                                               \
  }

#define CDSSTRESS_SequentialTreiberStack(test_fixture)                         \
  CDSSTRESS_SequentialTreiberStack_F(test_fixture, Treiber_HP);                \
  CDSSTRESS_SequentialTreiberStack_F(test_fixture, Treiber_DHP);               \

#define CDSSTRESS_SequentialEliminationStack(test_fixture)                     \
  CDSSTRESS_SequentialEliminationStack_F(test_fixture, Elimination_HP);        \
  CDSSTRESS_SequentialEliminationStack_F(test_fixture, Elimination_HP_dyn);    \
  CDSSTRESS_SequentialEliminationStack_F(test_fixture, Elimination_DHP);       \
  CDSSTRESS_SequentialEliminationStack_F(test_fixture, Elimination_DHP_dyn)

static size_t s_nSequentialStackPushCount = 100000;
static size_t s_nSequentialEliminationSize = 4;

class sequential_stack : public cds_test::stress_fixture {
protected:
  struct value_type {
    size_t nNo;
    size_t nThread;

    value_type() : nNo(0), nThread(0) {}
    value_type(size_t n) : nNo(n), nThread(0) {}
  };

protected:
  static void SetUpTestCase() {
    cds_test::config const &cfg = get_config("SequentialStack");
    GetConfig(SequentialStackPushCount);
    GetConfig(SequentialEliminationSize);
  }

  template <typename Stack> void test(Stack &stack) {
    size_t push_error_cnt = 0;
    size_t pop_sum = 0;
    value_type v;
    size_t m_nStartItem = 0;
    size_t m_nEndItem = s_nSequentialStackPushCount;
    for (v.nNo = m_nStartItem; v.nNo < m_nEndItem; ++v.nNo) {
      if (!stack.push(v))
        ++push_error_cnt;
    }

    while (stack.pop(v)) {
      pop_sum += v.nNo;
    }

    if (push_error_cnt) {
      std::cout << "Sequential stack push error count: " << push_error_cnt
                << "\n";
    }
    size_t supposed_sum =
        (m_nStartItem + m_nEndItem - 1) * (m_nEndItem - m_nStartItem) / 2;
    if (pop_sum != supposed_sum) {
      std::cout << "Sequential stack pop sum: " << pop_sum
                << " != " << supposed_sum << "\n";
    }
  }
};

CDSSTRESS_SequentialTreiberStack(sequential_stack)
    CDSSTRESS_SequentialEliminationStack(sequential_stack)

} // namespace
